/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_TASK_INTERNAL_TYPES_H
#define MARS_TASK_INTERNAL_TYPES_H

#include <stdint.h>

#include "mars/workload_types.h"

#define MARS_TASK_CONTEXT_SIZE				128
#define MARS_TASK_CONTEXT_ALIGN				16
#define MARS_TASK_CONTEXT_SAVE_ALIGN			128
#define MARS_TASK_CONTEXT_SAVE_UNIT_SIZE		8
#define MARS_TASK_CONTEXT_SAVE_UNIT_ALIGN		128
#define MARS_TASK_CONTEXT_SAVE_UNIT_ADDR_ALIGN_MASK	0xf
#define MARS_TASK_CONTEXT_SAVE_UNIT_SIZE_ALIGN_MASK	0xf

struct mars_task_context {
	struct mars_workload_module module;	/* workload module */
	int32_t exit_code;			/* exit code */
	uint64_t exec_ea;			/* ea of exec */
	uint32_t exec_size;			/* text/data size of exec */
	uint32_t bss_size;			/* bss size of exec */
	uint32_t entry;				/* entry address of exec */
	uint32_t stack;				/* stack pointer of exec */
	uint64_t context_save_area_ea;		/* context save area */
	uint64_t context_save_unit_ea;		/* context save unit list */
	struct mars_task_id id;			/* task id */
	struct mars_task_args args;		/* task args */
} __attribute__((aligned(MARS_TASK_CONTEXT_ALIGN)));

#endif
