/*
 * Copyright (C) 2008 Sony Computer Entertainment Inc.
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <mars/mars.h>
#include <spu_mfcio.h>
#include <mars_task_test.h>
#include "cmd.h"

#define FAST_DMA_ALIGN	128

static int32_t cmd __attribute__((aligned(FAST_DMA_ALIGN)));
static int32_t res __attribute__((aligned(FAST_DMA_ALIGN)));

int mars_task_main(const struct mars_task_args *task_args)
{
	int ret;
	uint64_t cmd_ea = task_args->type.u64[0];
	uint64_t res_ea = task_args->type.u64[1];

	for (;;) {
		mfc_get(&cmd, cmd_ea, sizeof(int32_t), 0, 0, 0);
		mfc_write_tag_mask(1);
		mfc_read_tag_status_all();

		if (cmd == CMD_NOP)
			continue;
		if (cmd == CMD_EXIT)
			break;
		if (cmd == CMD_YIELD) {
			ret = mars_task_yield();
			MARS_TEST_ASSERT_EQUAL(ret, MARS_SUCCESS);
		} else {
			MARS_TEST_ASSERT_EQUAL(cmd, CMD_HELLO);
			res = 1;
			mfc_put(&res, res_ea, sizeof(int32_t), 0, 0, 0);
			mfc_write_tag_mask(1);
			mfc_read_tag_status_all();
		}
	}

	return 0;
}
