/*
 * Copyright 2008 Sony Corporation
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <pthread.h>
#include <unistd.h>
#include <sys/time.h>
#include <libspe2.h>
#include <mars/mars.h>
#include <mars_test.h>

#define MAX_TASK_COUNT	256
#define VALID_TIME	10000

extern spe_program_handle_t mpu1_prog, mpu2_prog;

static struct mars_context *mars;
static struct mars_task_id task_ids[MAX_TASK_COUNT];
static struct mars_task_args task_args;
static pthread_t threads[2];
static long long times[2];
static struct mars_task_args triggers[2] __attribute__((aligned(16)));

long long int get_ticks()
{
  struct timeval tv;

  gettimeofday(&tv, NULL);
  return (long long int)tv.tv_sec*1000000 + tv.tv_usec;
}

void *thread_proc(void *arg)
{
	mars_test_counter_wait(&triggers[(int)(intptr_t)arg].type.u32[0], 1);

	times[(int)(intptr_t)arg] = get_ticks();

	return NULL;
}

int main(void)
{
	int i, exit_code, ret;
	uint32_t tb_freq;

	int spe_cnt = spe_cpu_info_get(SPE_COUNT_PHYSICAL_SPES, -1);

	MARS_TEST_ASSERT(spe_cnt <= MAX_TASK_COUNT);

	ret = mars_test_get_timebase_freq(&tb_freq);
	MARS_TEST_ASSERT_EQUAL(ret, 0);

	ret = mars_context_create(&mars, spe_cnt, 0);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	for (i = 0; i < spe_cnt; i++) {
		ret = mars_task_create(
			mars,
			&task_ids[i],
			"18002",
			(i < spe_cnt-1) ?
				mpu1_prog.elf_image :
				mpu2_prog.elf_image,
			MARS_TASK_CONTEXT_SAVE_ALL);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
	}

	triggers[0].type.u32[0] = 0;
	triggers[1].type.u32[0] = 0;

	pthread_create(&threads[0], NULL, thread_proc, (void *)(intptr_t)0);
	pthread_create(&threads[1], NULL, thread_proc, (void *)(intptr_t)1);

	task_args.type.u32[0] = tb_freq;
	task_args.type.u64[1] = mars_ptr_to_ea(&triggers[0]);
	task_args.type.u64[2] = mars_ptr_to_ea(&triggers[1]);

	for (i = 0; i < spe_cnt; i++)
		ret = mars_task_schedule(&task_ids[i], &task_args, 0);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	for (i = 0; i < spe_cnt; i++) {
		ret = mars_task_wait(&task_ids[i], &exit_code);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		MARS_TEST_ASSERT_EQUAL(exit_code, 0);
	}

	pthread_join(threads[0], NULL);
	pthread_join(threads[1], NULL);

	for (i = 0; i < spe_cnt; i++) {
		ret = mars_task_destroy(&task_ids[i]);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
	}

	ret = mars_context_destroy(mars);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	MARS_TEST_ASSERT(times[1]-times[0] < VALID_TIME);

	return 0;
}
