/*
 * Copyright 2008 Sony Corporation
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <unistd.h>
#include <mars/mars.h>
#include <libspe2.h>
#include <mars_test.h>

#define WAIT_NUM (MARS_TASK_BARRIER_WAIT_MAX / 2)
#define NUM_TASKS (WAIT_NUM * 2)

extern spe_program_handle_t mpu_prog;

static struct mars_context *mars_ctx;
static struct mars_task_id task_id[NUM_TASKS];
static struct mars_task_args task_args[NUM_TASKS];
static uint64_t barrier;
static uint32_t notify[32] __attribute((aligned(128)));
static uint32_t shared_ret[NUM_TASKS * 4] __attribute((aligned(16)));


int main(void)
{
	int ret, count, num_success;

	ret = mars_context_create(&mars_ctx, 0, 0);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	notify[0] = 0;

	ret = mars_task_barrier_create(mars_ctx,
		&barrier, WAIT_NUM);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	/* create tasks and notify */
	for (count = 0; count < NUM_TASKS; count++) {
		shared_ret[count * 4] = 0;
		ret = mars_task_create(mars_ctx, &task_id[count], NULL,
		mpu_prog.elf_image, MARS_TASK_CONTEXT_SAVE_ALL);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		task_args[count].type.u64[0] = barrier;
		task_args[count].type.u64[1] =
			mars_ptr_to_ea(&notify);
		task_args[count].type.u64[2] =
			mars_ptr_to_ea(&shared_ret[count * 4]);
		ret = mars_task_schedule(
				&task_id[count], &task_args[count], 0);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
	}

	sleep(1);
	/* mars_task_barrier_notify at the same time */
	mars_test_counter_set(&notify[0], 1);
	sleep(1);

	/* check notify count */
	num_success = 0;
	for (count = 0; count < NUM_TASKS; count++) {
		if (shared_ret[count * 4] == 1)
			num_success++;
	}
	MARS_TEST_ASSERT_EQUAL(num_success, WAIT_NUM);

	/* blocking MPU tasks can not be destroyed by host */
	exit(0);
}

