/*
 * Copyright 2008 Sony Corporation
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <unistd.h>
#include <mars/mars.h>
#include <libspe2.h>
#include <mars_test.h>

#define NUM_TASKS MARS_TASK_BARRIER_WAIT_MAX

extern spe_program_handle_t mpu1_prog;
extern spe_program_handle_t mpu2_prog;

#define CHK_NUM 2
static uint32_t chk_num[] = {
	NUM_TASKS / 2,
	NUM_TASKS
};

static struct mars_context *mars_ctx;
static struct mars_task_id task1_id[NUM_TASKS], task2_id;
static struct mars_task_args task1_args, task2_args;
static uint64_t barrier[CHK_NUM];
static uint32_t shared_resource __attribute((aligned(16)));

int main(void)
{
	int ret, i, count;
	int32_t exit_code;

	ret = mars_context_create(&mars_ctx, 0, 0);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	for (i = 0; i < CHK_NUM; i++) {
		shared_resource = 0;

		ret = mars_task_barrier_create(mars_ctx,
			&barrier[i], chk_num[i]);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* run notify/wait tasks */
		for (count = 0; count < chk_num[i] - 1; count++) {
			ret = mars_task_create(mars_ctx,
				&task1_id[count], NULL,
				mpu1_prog.elf_image,
				MARS_TASK_CONTEXT_SAVE_ALL);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

			task1_args.type.u64[0] = barrier[i];
			task1_args.type.u64[1] =
				mars_ptr_to_ea(&shared_resource);
			ret = mars_task_schedule(&task1_id[count],
				&task1_args, 0);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		}

		/* check if mpu task is waiting */
		sleep(1);
		MARS_TEST_ASSERT_EQUAL(shared_resource, 0);

		/* notify */
		ret = mars_task_create(mars_ctx, &task2_id, NULL,
			mpu2_prog.elf_image, MARS_TASK_CONTEXT_SAVE_ALL);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		task2_args.type.u64[0] = barrier[i];
		ret = mars_task_schedule(&task2_id, &task2_args, 0);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		ret = mars_task_wait(&task2_id, &exit_code);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		MARS_TEST_ASSERT_ERROR(exit_code, MARS_SUCCESS);

		sleep(1);
		/* check if mpu tasks still run */
		MARS_TEST_ASSERT_EQUAL(shared_resource, 1);
		for (count = 0; count < chk_num[i] - 1; count++) {
			ret = mars_task_try_wait(&task1_id[count], &exit_code);
			MARS_TEST_ASSERT_ERROR(ret, MARS_ERROR_BUSY);
		}

		/* 2nd notify */
		task2_args.type.u64[0] = barrier[i];
		ret = mars_task_schedule(&task2_id, &task2_args, 0);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		ret = mars_task_wait(&task2_id, &exit_code);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		MARS_TEST_ASSERT_ERROR(exit_code, MARS_SUCCESS);

		for (count = 0; count < chk_num[i] - 1; count++) {
			ret = mars_task_wait(&task1_id[count], &exit_code);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
			MARS_TEST_ASSERT_ERROR(exit_code, MARS_SUCCESS);

			ret = mars_task_destroy(&task1_id[count]);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		}

		ret = mars_task_destroy(&task2_id);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		ret = mars_task_barrier_destroy(barrier[i]);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
	}

	ret = mars_context_destroy(mars_ctx);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	return 0;
}

