/*
 * Copyright 2008 Sony Corporation
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <unistd.h>
#include <mars/mars.h>
#include <libspe2.h>
#include <mars_test.h>

#define NUM_TASKS MARS_TASK_BARRIER_WAIT_MAX

extern spe_program_handle_t mpu_prog;

static struct mars_context *mars_ctx;
static struct mars_task_id task_id[NUM_TASKS];
static struct mars_task_args task_args[NUM_TASKS];
static uint64_t barrier[NUM_TASKS];
static uint32_t notify[32] __attribute((aligned(128)));

#define CHK_NUM 5
static uint32_t chk_num[] = {
	1,
	NUM_TASKS / 2,
	NUM_TASKS,
	NUM_TASKS,
	NUM_TASKS
};

int main(void)
{
	int ret, i, count, num_success;
	int32_t exit_code;

	ret = mars_context_create(&mars_ctx, 0, 0);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	for (i = 0; i < CHK_NUM; i++) {
		notify[0] = 0;

		ret = mars_task_barrier_create(mars_ctx,
			&barrier[i], chk_num[i]);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* create tasks and notify */
		for (count = 0; count < NUM_TASKS; count++) {
			ret = mars_task_create(mars_ctx, &task_id[count], NULL,
			mpu_prog.elf_image, MARS_TASK_CONTEXT_SAVE_ALL);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

			task_args[count].type.u64[0] = barrier[i];
			task_args[count].type.u64[1] =
				mars_ptr_to_ea(&notify);
			ret = mars_task_schedule(
					&task_id[count], &task_args[count], 0);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		}

		sleep(1);
		/* mars_task_barrier_notify at the same time */
		mars_test_counter_set(&notify[0], 1);
		sleep(1);

		/* check */
		num_success = 0;
		for (count = 0; count < NUM_TASKS; count++) {
			ret = mars_task_wait(&task_id[count], &exit_code);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
			if (exit_code == MARS_SUCCESS)
				num_success++;
			else if (exit_code != MARS_ERROR_BUSY)
				MARS_TEST_ASSERT_EQUAL(exit_code, 0);
		}

		MARS_TEST_ASSERT_EQUAL(num_success, chk_num[i]);

		for (count = 0; count < NUM_TASKS; count++) {

			ret = mars_task_destroy(&task_id[count]);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		}
		ret = mars_task_barrier_destroy(barrier[i]);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
	}

	ret = mars_context_destroy(mars_ctx);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	return 0;
}

