/*
 * Copyright 2008 Sony Corporation
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <unistd.h>
#include <mars/mars.h>
#include <libspe2.h>
#include <mars_test.h>

#define EVENT_FLAG_SIZE 32

extern spe_program_handle_t mpu_prog;
#define NUM_PATTERNS 5
static uint32_t pattern[NUM_PATTERNS] = {
	0x00000001UL,
	0x0000CAFEUL,
	0x01010101UL,
	0x5A5A5A5AUL,
	0xFFFFFFFFUL
};
static struct mars_context *mars_ctx;
static struct mars_task_id task_id[EVENT_FLAG_SIZE];
static struct mars_task_args task_args;
static uint64_t mpu_to_host[NUM_PATTERNS];
static uint32_t notify[32] __attribute((aligned(128)));


static void check_flag(int id, uint32_t chk_bit)
{
	int ret;
	uint32_t val = 0;

	ret = mars_task_event_flag_try_wait(
			mpu_to_host[id],
			chk_bit,
			MARS_TASK_EVENT_FLAG_MASK_OR, &val);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	MARS_TEST_ASSERT_EQUAL(chk_bit, val);
}

int main(void)
{
	int ret, i, bits;
	uint32_t chk_bit, pat;
	int32_t exit_code;

	ret = mars_context_create(&mars_ctx, 0, 0);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	for (i = 0; i < NUM_PATTERNS; i++) {
		notify[0] = 0;
		ret = mars_task_event_flag_create(
			mars_ctx,
			&mpu_to_host[i],
			MARS_TASK_EVENT_FLAG_MPU_TO_HOST,
			MARS_TASK_EVENT_FLAG_CLEAR_AUTO);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		chk_bit = 0x01;
		pat = pattern[i];
		for (bits = 0; pat; bits++, pat >>= 1, chk_bit <<= 1) {
			if (!(pat & 0x01))
				continue;
			ret = mars_task_create(mars_ctx, &task_id[bits], NULL,
				mpu_prog.elf_image,
				MARS_TASK_CONTEXT_SAVE_ALL);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

			/* mpu tasks will set event_flags */
			task_args.type.u64[0] = mpu_to_host[i];
			task_args.type.u64[1] = (uint64_t)chk_bit;
			task_args.type.u64[2] = mars_ptr_to_ea(&notify);
			ret = mars_task_schedule(&task_id[bits], &task_args, 0);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		}

		sleep(1);
		/* start all tasks */
		mars_test_counter_set(&notify[0], 1);

		pat = pattern[i];
		for (bits = 0; pat; bits++, pat >>= 1) {
			if (!(pat & 0x01))
				continue;
			ret = mars_task_wait(&task_id[bits], &exit_code);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
			MARS_TEST_ASSERT_ERROR(exit_code, MARS_SUCCESS);

			ret = mars_task_destroy(&task_id[bits]);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		}

		check_flag(i, pattern[i]);

		ret = mars_task_event_flag_destroy(mpu_to_host[i]);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
	}

	ret = mars_context_destroy(mars_ctx);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	return 0;
}

