/*
 * Copyright 2008 Sony Corporation
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <pthread.h>
#include <unistd.h>
#include <mars/mars.h>
#include <libspe2.h>
#include <mars_test.h>

#define FLAG_WAIT_USEC 10000 /* 10ms */

#define EVENT_FLAG_SIZE 32

extern spe_program_handle_t mpu_prog;
#define NUM_PATTERNS 3
static uint32_t pattern[NUM_PATTERNS][2] = {
/*        pattern       or chk_bit   */
	{ 0x00000001UL, 0x00000001UL },
	{ 0x0000CAFEUL, 0x00008000UL },
	{ 0x10203040UL, 0x10000000UL }
};
static struct mars_context *mars_ctx;
static struct mars_task_id task_id;
static struct mars_task_args task_args, thread_args;
static uint64_t mpu_to_host[NUM_PATTERNS];
static uint32_t notify[32] __attribute((aligned(128)));

static pthread_t thread;

static void *thread1_proc(void *ptr)
{
	int ret;
	struct mars_task_args *task_args = (struct mars_task_args *)ptr;
	uint32_t i = (uint32_t)task_args->type.u32[0];
	uint32_t chk_bit = task_args->type.u32[1];

	mars_test_counter_wait(&notify[0], 2);

	ret = mars_task_event_flag_wait(
			mpu_to_host[i],
			chk_bit,
			MARS_TASK_EVENT_FLAG_MASK_AND, NULL);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	mars_test_counter_set(&notify[0], 3);

	return NULL;
}
static void *thread2_proc(void *ptr)
{
	int ret;
	struct mars_task_args *task_args = (struct mars_task_args *)ptr;
	uint32_t i = (uint32_t)task_args->type.u32[0];
	uint32_t chk_bit = task_args->type.u32[1];

	mars_test_counter_wait(&notify[0], 2);

	ret = mars_task_event_flag_wait(
			mpu_to_host[i],
			chk_bit,
			MARS_TASK_EVENT_FLAG_MASK_OR, NULL);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	mars_test_counter_set(&notify[0], 3);

	return NULL;
}

int main(void)
{
	int ret, i;
	int32_t exit_code;

	ret = mars_context_create(&mars_ctx, 0, 0);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	/* wait condition AND */
	for (i = 0; i < NUM_PATTERNS; i++) {
		notify[0] = 0;
		ret = mars_task_event_flag_create(
			mars_ctx,
			&mpu_to_host[i],
			MARS_TASK_EVENT_FLAG_MPU_TO_HOST,
			MARS_TASK_EVENT_FLAG_CLEAR_AUTO);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		thread_args.type.u32[0] = (uint32_t)i;
		thread_args.type.u32[1] = (uint32_t)pattern[i][0];
		pthread_create(&thread, NULL, thread1_proc, &thread_args);

		/* create mpu task */
		ret = mars_task_create(mars_ctx, &task_id, NULL,
			mpu_prog.elf_image,
			MARS_TASK_CONTEXT_SAVE_ALL);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* mpu tasks will set event_flags for and wait condition */
		task_args.type.u64[0] = mpu_to_host[i];
		task_args.type.u64[1] = (uint64_t)pattern[i][0];
		task_args.type.u64[2] = mars_ptr_to_ea(&notify);
		ret = mars_task_schedule(&task_id, &task_args, 0);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* start mars_task_event_flag_set by mpu */
		mars_test_counter_set(&notify[0], 1);

		/* wait mars_task_event_flag_wait done */
		usleep(FLAG_WAIT_USEC);
		/* if notify is 3, mars_task_event_flag_wait is done */
		MARS_TEST_ASSERT_EQUAL(notify[0], 3);

		ret = mars_task_wait(&task_id, &exit_code);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		MARS_TEST_ASSERT_ERROR(exit_code, MARS_SUCCESS);

		ret = mars_task_destroy(&task_id);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		pthread_join(thread, NULL);

		ret = mars_task_event_flag_destroy(mpu_to_host[i]);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
	}
	/* wait condition OR */
	for (i = 0; i < NUM_PATTERNS; i++) {
		notify[0] = 0;
		ret = mars_task_event_flag_create(
			mars_ctx,
			&mpu_to_host[i],
			MARS_TASK_EVENT_FLAG_MPU_TO_HOST,
			MARS_TASK_EVENT_FLAG_CLEAR_AUTO);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		thread_args.type.u32[0] = (uint32_t)i;
		thread_args.type.u32[1] = (uint32_t)pattern[i][0];
		pthread_create(&thread, NULL, thread2_proc, &thread_args);

		/* create mpu task */
		ret = mars_task_create(mars_ctx, &task_id, NULL,
			mpu_prog.elf_image,
			MARS_TASK_CONTEXT_SAVE_ALL);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* mpu tasks will set event_flag for or wait condition */
		task_args.type.u64[0] = mpu_to_host[i];
		task_args.type.u64[1] = (uint64_t)pattern[i][1];
		task_args.type.u64[2] = mars_ptr_to_ea(&notify);
		ret = mars_task_schedule(&task_id, &task_args, 0);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* start mars_task_event_flag_set by mpu */
		mars_test_counter_set(&notify[0], 1);

		/* wait mars_task_event_flag_wait done */
		usleep(FLAG_WAIT_USEC);
		/* if notify is 3, mars_task_event_flag_wait is done */
		MARS_TEST_ASSERT_EQUAL(notify[0], 3);

		ret = mars_task_wait(&task_id, &exit_code);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		MARS_TEST_ASSERT_ERROR(exit_code, MARS_SUCCESS);

		ret = mars_task_destroy(&task_id);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		pthread_join(thread, NULL);

		ret = mars_task_event_flag_destroy(mpu_to_host[i]);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
	}

	ret = mars_context_destroy(mars_ctx);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	return 0;
}

