/*
 * Copyright 2008 Sony Corporation
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <unistd.h>
#include <mars/mars.h>
#include <libspe2.h>
#include <mars_test.h>

#define FLAG_WAIT_USEC 10000 /* 10ms */

#define EVENT_FLAG_SIZE 32
extern spe_program_handle_t mpu1_prog;
extern spe_program_handle_t mpu2_prog;
extern spe_program_handle_t mpu3_prog;
#define NUM_PATTERNS 3
static uint32_t pattern[NUM_PATTERNS][2] = {
/*        pattern       or chk_bit   */
	{ 0x00000001UL, 0x00000001UL },
	{ 0x0000CAFEUL, 0x00008000UL },
	{ 0x10203040UL, 0x10000000UL }
};
static struct mars_context *mars_ctx;
static struct mars_task_id task_id, task3_id;
static struct mars_task_args task_args, task3_args;
static uint64_t mpu_to_mpu[NUM_PATTERNS];
static uint32_t notify[32] __attribute((aligned(128)));

int main(void)
{
	int ret, i;
	int32_t exit_code;

	ret = mars_context_create(&mars_ctx, 0, 0);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	/* wait conditon AND */
	for (i = 0; i < NUM_PATTERNS; i++) {
		notify[0] = 0;
		ret = mars_task_event_flag_create(
			mars_ctx,
			&mpu_to_mpu[i],
			MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
			MARS_TASK_EVENT_FLAG_CLEAR_AUTO);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		ret = mars_task_create(mars_ctx, &task_id, NULL,
			mpu1_prog.elf_image,
			MARS_TASK_CONTEXT_SAVE_ALL);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		task_args.type.u64[0] = mpu_to_mpu[i];
		task_args.type.u64[1] = (uint64_t)pattern[i][0];
		task_args.type.u64[2] = mars_ptr_to_ea(&notify);

		ret = mars_task_schedule(&task_id, &task_args, 0);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* create flag_set task */
		ret = mars_task_create(mars_ctx, &task3_id, NULL,
			mpu3_prog.elf_image,
			MARS_TASK_CONTEXT_SAVE_ALL);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* the tasks will set event_flags for and wait condition */
		task3_args.type.u64[0] = mpu_to_mpu[i];
		task3_args.type.u64[1] = (uint64_t)pattern[i][0];
		task3_args.type.u64[2] = mars_ptr_to_ea(&notify);
		ret = mars_task_schedule(&task3_id, &task3_args, 0);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* start mars_task_event_flag_set by mpu */
		mars_test_counter_set(&notify[0], 1);

		/* wait mars_task_event_flag_wait done */
		usleep(FLAG_WAIT_USEC);
		MARS_TEST_ASSERT_EQUAL(notify[0], 3);

		ret = mars_task_wait(&task_id, &exit_code);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		MARS_TEST_ASSERT_ERROR(exit_code, MARS_SUCCESS);

		ret = mars_task_destroy(&task_id);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		ret = mars_task_wait(&task3_id, &exit_code);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		MARS_TEST_ASSERT_ERROR(exit_code, MARS_SUCCESS);

		ret = mars_task_destroy(&task3_id);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		ret = mars_task_event_flag_destroy(mpu_to_mpu[i]);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
	}
	/* wait conditon OR */
	for (i = 0; i < NUM_PATTERNS; i++) {
		notify[0] = 0;
		ret = mars_task_event_flag_create(
			mars_ctx,
			&mpu_to_mpu[i],
			MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
			MARS_TASK_EVENT_FLAG_CLEAR_AUTO);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		ret = mars_task_create(mars_ctx, &task_id, NULL,
			mpu2_prog.elf_image,
			MARS_TASK_CONTEXT_SAVE_ALL);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		task_args.type.u64[0] = mpu_to_mpu[i];
		task_args.type.u64[1] = (uint64_t)pattern[i][0];
		task_args.type.u64[2] = mars_ptr_to_ea(&notify);

		ret = mars_task_schedule(&task_id, &task_args, 0);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* create flag_set task */
		ret = mars_task_create(mars_ctx, &task3_id, NULL,
			mpu3_prog.elf_image,
			MARS_TASK_CONTEXT_SAVE_ALL);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* the tasks will set event_flag for or wait condition */
		task3_args.type.u64[0] = mpu_to_mpu[i];
		task3_args.type.u64[1] = (uint64_t)pattern[i][1];
		task3_args.type.u64[2] = mars_ptr_to_ea(&notify);
		ret = mars_task_schedule(&task3_id, &task3_args, 0);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* start mars_task_event_flag_set by mpu */
		mars_test_counter_set(&notify[0], 1);


		/* wait mars_task_event_flag_wait done */
		usleep(FLAG_WAIT_USEC);
		MARS_TEST_ASSERT_EQUAL(notify[0], 3);

		ret = mars_task_wait(&task_id, &exit_code);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		MARS_TEST_ASSERT_ERROR(exit_code, MARS_SUCCESS);

		ret = mars_task_destroy(&task_id);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		ret = mars_task_wait(&task3_id, &exit_code);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		MARS_TEST_ASSERT_ERROR(exit_code, MARS_SUCCESS);

		ret = mars_task_destroy(&task3_id);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		ret = mars_task_event_flag_destroy(mpu_to_mpu[i]);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
	}

	ret = mars_context_destroy(mars_ctx);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	return 0;
}

