/*
 * Copyright 2008 Sony Corporation
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <unistd.h>
#include <mars/mars.h>
#include <libspe2.h>
#include <mars_test.h>

#define TASK_MAX MARS_TASK_EVENT_FLAG_WAIT_MAX

#define EVENT_FLAG_SIZE 32
extern spe_program_handle_t mpu1_prog;
extern spe_program_handle_t mpu2_prog;
#define NUM_PATTERNS 5
static uint32_t pattern[NUM_PATTERNS] = {
	0x00000001UL,
	0x0000CAFEUL,
	0x10203040UL,
	0x5A5A5A5AUL,
	0xFFFFFFFFUL
};
static struct mars_context *mars_ctx;
static struct mars_task_id task_id[TASK_MAX], task2_id;
static struct mars_task_args task_args, task2_args;
static uint64_t mpu_to_mpu[NUM_PATTERNS];
static uint32_t shared_resource[TASK_MAX * 4] __attribute((aligned(16)));

int main(void)
{
	int ret, i, count;
	int32_t exit_code;

	ret = mars_context_create(&mars_ctx, 0, 0);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	for (i = 0; i < NUM_PATTERNS; i++) {
		ret = mars_task_event_flag_create(
			mars_ctx,
			&mpu_to_mpu[i],
			MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
			MARS_TASK_EVENT_FLAG_CLEAR_AUTO);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		/* create wait mpu tasks */
		for (count = 0; count < TASK_MAX; count++) {
			shared_resource[count * 4] = 0;
			ret = mars_task_create(mars_ctx, &task_id[count], NULL,
				mpu1_prog.elf_image,
				MARS_TASK_CONTEXT_SAVE_ALL);
				MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

			task_args.type.u64[0] = mpu_to_mpu[i];
			task_args.type.u64[1] = (uint64_t)pattern[i];
			task_args.type.u64[2] =
				mars_ptr_to_ea(&shared_resource[count * 4]);

			ret = mars_task_schedule(&task_id[count],
				&task_args, 0);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		}

		sleep(1);
		/* check if all tasks are waiting */
		for (count = 0; count < TASK_MAX; count++)
			MARS_TEST_ASSERT_EQUAL(shared_resource[count * 4], 0);

		/* create set flag task */
		ret = mars_task_create(mars_ctx, &task2_id, NULL,
			mpu2_prog.elf_image,
			MARS_TASK_CONTEXT_SAVE_ALL);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		task2_args.type.u64[0] = mpu_to_mpu[i];
		task2_args.type.u64[1] = (uint64_t)pattern[i];
		ret = mars_task_schedule(&task2_id, &task2_args, 0);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		ret = mars_task_wait(&task2_id, &exit_code);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		MARS_TEST_ASSERT_ERROR(exit_code, MARS_SUCCESS);

		/* check if all waiting condition is satisfied */
		for (count = 0; count < TASK_MAX; count++) {
			ret = mars_task_wait(&task_id[count], &exit_code);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
			MARS_TEST_ASSERT_ERROR(exit_code, MARS_SUCCESS);

			ret = mars_task_destroy(&task_id[count]);
			MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
		}

		ret = mars_task_destroy(&task2_id);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

		ret = mars_task_event_flag_destroy(mpu_to_mpu[i]);
		MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);
	}

	ret = mars_context_destroy(mars_ctx);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	return 0;
}

