/*
 * Copyright 2008 Sony Corporation
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <unistd.h>
#include <mars/mars.h>
#include <libspe2.h>
#include <mars_test.h>

#define SEMAPHORE_COUNTS 10

#define WAIT_SEC 2
extern spe_program_handle_t mpu_prog;

static struct mars_context *mars_ctx;
static struct mars_task_id task_id;
static struct mars_task_args task_args;
static uint64_t semaphore;
static uint32_t shared_resource __attribute__((aligned(16)));

int main(void)
{
	int ret;

	ret = mars_context_create(&mars_ctx, 0, 0);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	ret = mars_task_semaphore_create(
		mars_ctx,
		&semaphore,
		SEMAPHORE_COUNTS);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	ret = mars_task_create(mars_ctx, &task_id, NULL,
		mpu_prog.elf_image, MARS_TASK_CONTEXT_SAVE_ALL);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	shared_resource = 0;
	task_args.type.u64[0] = semaphore;
	task_args.type.u64[1] = mars_ptr_to_ea(&shared_resource);
	task_args.type.u64[2] = SEMAPHORE_COUNTS;
	ret = mars_task_schedule(&task_id, &task_args, 0);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	/* wait for blocking */
	sleep(WAIT_SEC);

	/* Check if shared_resource hasn't been changed  */
	MARS_TEST_ASSERT_EQUAL(shared_resource, 0);

	ret = mars_task_semaphore_destroy(semaphore);
	MARS_TEST_ASSERT_ERROR(ret, MARS_SUCCESS);

	exit(0);
}

